#! /usr/bin/env python
# Copyright (c) 2018 - 2019 Jolle Jolles <j.w.jolles@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import math
import numpy as np

def uneven(value):

    """Returns the closest uneven value equal to or lower than provided"""

    if value == 0:
        newvalue = 1
    else:
        newvalue = value -1 if value % 2 == 0 else value

    return newvalue


def closenr(n, m) :

    """Find the number closest to n and divisible by m"""

    q = int(n / m)
    n1 = m * q
    if((n * m) > 0) :
        n2 = (m * (q + 1))
    else :
        n2 = (m * (q - 1))

    if (abs(n - n1) < abs(n - n2)) :
        return n1

    return n2


def seqcount(start, stop, steplength):

    """
    Returns a sequence of numbers between a start and stop value with a
    certain steplength
    """

    step = (stop - start) / steplength
    step = int(np.ceil(step))
    sequence = range(start, stop, step)

    return sequence


def maxsteps(value, maxval = 500):

    """
    Returns the maximum number of steps and stepsize for a value that may be
    divided up to a maximum value, all being rounded value. For example, if one
    wants to get the maximum number of steps to get to a value of 100 where it
    can be maximally divided 7 times, it will return nsteps=5, stepsize=20.
    """

    valrange = [value-3,value-2,value-1,value]
    nsteps = 1
    for val in valrange:
        for _,n in enumerate(reversed(list(range(1, maxval)))):
            if val % n == 0:
                if n > nsteps:
                    nsteps = int(n)
                    stepsize = int(val/nsteps)
                break

    return(nsteps, stepsize)


def get_weights(w = 1.7, length = 20):

    """Returns a list of weights, based on quadratic function"""

    return [w**i for i in range(length, 0, -1)]


def sort_twoPoint(coords):

    """Fixes the coordinates of a two-point to be LT to BR"""

    x1 = min(coords[0][0], coords[1][0])
    y1 = min(coords[0][1], coords[1][1])
    x2 = max(coords[0][0], coords[1][0])
    y2 = max(coords[0][1], coords[1][1])

    return ((x1,y1),(x2,y2))


def sort_points(pts):

    """
    Initializes a list of coordinates that will be ordered from top-left to
    bottom-left in clockwise order
    """

    rect = np.zeros((4, 2), dtype = "float32")
    s = pts.sum(axis = 1)
    rect[0] = pts[np.argmin(s)]
    rect[2] = pts[np.argmax(s)]
    diff = np.diff(pts, axis = 1)
    rect[1] = pts[np.argmin(diff)]
    rect[3] = pts[np.argmax(diff)]

    return rect


def points_to_vec(pt1, pt2, flip = False):

    """
    Converts the coordinate of two points (pt1 > pt2) to a vector

    flip: bool, default = False
        If the coordinate system should be flipped such that higher y-coords
        are lower (e.g. needed when working with images in opencv).
    """

    vx = pt2[0] - pt1[0]
    vy = pt1[1] - pt2[1] if flip else pt2[1] - pt1[1]

    return vx, vy


def angle_to_vec(angle):

    """
    Converts an angle in degrees to a vector. Uses a coordinate system that
    points north and ranges from -180 to 180 degrees.
    """

    vx = np.round(np.sin(np.radians(angle)), 3)
    vy = np.round(np.cos(np.radians(angle)), 3)

    return vx, vy


def points_to_angle(pt1, pt2 = None, flip = False):

    """
    Returns the angle of a vector from the origin to a single point or the angle
    between two points. Uses a coordinate system that points north and ranges
    from -180 to 180 degrees.

    flip: bool, default = False
        If the coordinate system should be flipped such that higher y-coords
        are lower (e.g. needed when working with images in opencv).
    """

    vx, vy = pt1 if pt2 is None else points_to_vec(pt1, pt2, flip)
    angle = np.round(np.arctan2(vx, vy) * 180 / np.pi,2)

    return angle


def midpoint(pt1, pt2):

    """Computes the midpoint between two points"""

    x = pt2[0]+int((pt1[0]-pt2[0])/2)
    y = pt2[1]+int((pt1[1]-pt2[1])/2)

    return (x,y)


def ptsToDist(pt1, pt2):

    """Computes the distance between two points"""

    if None in pt1 or None in pt2:
        dist = None
    else:
        vx, vy = points_to_vec(pt1, pt2)
        dist = np.linalg.norm([(vx, vy)])

    return dist


def diff_series(series, period = 1):

    """Shifts a pandas series upwards with a given period"""

    series2 = series.shift(periods = -period)

    return series2 - series


def dist_to_segment(pt, segment):

    """
    Calculates the distance between a point and a non-infinite line segment

    Explanation
    ----------
    To calculate the closest distance to a line segment, we first need to check
    if the point projects onto the line segment.  If it does, then we calculate
    the orthogonal distance from the point to the line. If the point does not
    project to the line segment, we calculate the distance to both endpoints
    and take the shortest distance.

    Input
    ----------
    point : tuple of the point's coordinates
    line : tuple of segment endpoints' coordinates

    Returns
    -------
    dist : minimum distance to the end point on the line
    coords : the relative x and y coordinates to the line
    """

    # unit vector
    segment = np.array(segment)
    uline = segment[1] - segment[0]
    norm_uline = uline / np.linalg.norm(uline)

    # compute the perpendicular distance to the theoretical infinite line
    dott_product = np.cross(segment[1] - segment[0], segment[0] - pt)
    dist_infline = (np.linalg.norm(dott_product / np.linalg.norm(uline)))
    diff = (norm_uline[0] * (pt[0] - segment[0][0])) + \
           (norm_uline[1] * (pt[1] - segment[0][1]))
    x_seg = np.round((norm_uline[0] * diff) + segment[0][0])
    y_seg = np.round((norm_uline[1] * diff) + segment[0][1])

    linept1dis = np.linalg.norm(segment[0] - pt)
    linept2dis = np.linalg.norm(segment[1] - pt)
    endpoint_dist = min(linept1dis, linept2dis)

    # decide if the intersection point falls on the line segment
    lp1_x = line[0][0]
    lp1_y = line[0][1]
    lp2_x = line[1][0]
    lp2_y = line[1][1]
    is_betw_x = lp1_x <= x_seg <= lp2_x or lp2_x <= x_seg <= lp1_x
    is_betw_y = lp1_y <= y_seg <= lp2_y or lp2_y <= y_seg <= lp1_y

    if is_betw_x and is_betw_y:
        dist = dist_infline
        coords = (x_seg, y_seg)
    else:
        # if not, then return the minimum distance to the segment endpoints
        dist = endpoint_dist
        coords = (line[0][0],line[0][1]) if linept1dis<=linept2dis else (line[1][0],line[1][1])

    return dist, coords


def maxrect(dims, maxdims = (1640,1232), decimals = 2):

    """
    Computes the maximal-sized rectangle with the same radio of dimensions as
    the rectangle provided, within the maximum dimensions
    """

    ratio = float(dims[1])/dims[0]
    maxpix = maxdims[0]*maxdims[1]
    w = round(math.sqrt(maxpix/ratio),decimals)
    h = round(w*ratiod,decimals)

    return (w,h)



#! /usr/bin/env python
# Copyright (c) 2018 - 2019 Jolle Jolles <j.w.jolles@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import io
import os
import re
import sys
import time
import socket
import fractions
from objsize import get_deep_size

class Suppressor(object):

    """Suppressed the output of a function"""
    def __enter__(self):
        self.stdout = sys.stdout
        sys.stdout = self

    def __exit__(self, type, value, traceback):
        sys.stdout = self.stdout
        if type is not None:
            print("exception")

    def write(self, x): pass


class Logger(object):

    """
    Class to log output of the command line to a log file
    """

    def __init__(self, filename):
        self.filename = filename

    class Transcript:
        def __init__(self, filename):
            self.terminal = sys.stdout
            self.log = open(filename, "a")
        def __getattr__(self, attr):
            return getattr(self.terminal, attr)
        def write(self, message):
            self.terminal.write(message)
            self.log.write(message)
        def flush(self):
            pass

    def start(self):
        sys.stdout = self.Transcript(self.filename)

    def stop(self):
        sys.stdout.log.close()
        sys.stdout = sys.stdout.terminal


class objectview:

    """Transforms dictionary into an object"""

    def __init__(self, dic):
        self.__dict__ = dic


def get_size(object):

    """Gets the size in bytes of an object"""

    size_bytes = get_deep_size(object)
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)

    return "%s %s" % (s, size_name[i])


def removeline(linenr=1):

    """Removes printed lines in terminal. Linenr starts with current line"""

    for _ in range(linenr):
        sys.stdout.write('\x1b[1A')
        sys.stdout.write('\x1b[2K')


def lineprint(text, stamp=True, newline=True, **kwargs):

    """
    Print text with simple timestap (stamp=True) and hostname (label=XXX) added
    """

    global line, label

    def _vardefined(var):

        return var in [var for var,_ in globals().items()]

    line = line if _vardefined("line") else ""
    label = label if _vardefined("label") else socket.gethostname()
    if not _vardefined("label"):
        label = ""

    if "label" in kwargs:
        label = kwargs["label"]

    if stamp:
        text = time.strftime("%H:%M:%S") + " [" + label + "] - " + text

    if newline:
        print(text)
    else:
        print(text, end=" ")


def homedir():

    """Returns the home directory"""

    return os.path.expanduser("~")+"/"


def isscript():

    """Determines if session is script or interactive (jupyter)"""

    import __main__ as main
    return hasattr(main, '__file__')


def isrpi(message=False):

    """Checks if current system is a Raspberry Pi"""

    try:
        with io.open('/proc/cpuinfo', 'r') as cpuinfo:
            found = False
            for line in cpuinfo:
                if line.startswith('Hardware'):
                    found = True
                    label, value = line.strip().split(':', 1)
                    value = value.strip()
                    if value not in ('BCM2708','BCM2709','BCM2835','BCM2836',
                                     'BCM2837','BCM2837B0','BCM2711'):
                        return False
            if not found:
                raise ValueError("""Unable to determine if system is rpi or
                                 not. Set system manually""")

    except IOError:
        if message:
            lineprint("non-rpi system detected..")
        return False

    if message:
        lineprint("rpi system detected..")
    return True


def checkfrac(input):

    """Checks string for Fractions and converts them accordingly"""

    transformed_text = re.sub(r'([\d.]+)', r'fractions.Fraction("\1")', input)

    return eval(transformed_text)



#! /usr/bin/env python
# Copyright (c) 2018 - 2019 Jolle Jolles <j.w.jolles@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division
from __future__ import print_function

import os
import cv2
import numpy as np

from pythutils.fileutils import get_ext
from pythutils.mathutils import closenr, sort_points


def check_media(source, internal=False):

    """Runs some basic checks on a mediafile or stream"""

    ext = get_ext(str(source))
    ftype = None
    if ext in [".mov",".mp4",".avi"]:
        ftype = "vid"
    if ext in [".jpg", ".png", ".jpeg", ".bmp"]:
        ftype = "img"
    if type(source) == int:
        ftype = "stream"
    if ftype == None:
        print("File neither video or image file..")
        return False

    if ftype == "img" or ftype == "vid":
        filedir = os.path.dirname(source)
        if filedir != "":
            if not os.path.isdir(filedir):
                print("File directory does not exist..")
                return False
        if not os.path.isfile(source):
            print("File does not exist..")
            return False

    if ftype == "vid" or ftype == "stream":
        cap = cv2.VideoCapture(source)
        flag, frame = cap.read()
        if not flag:
            print("Video source opened but failed to read images..")
            return False

    if not internal:
        print("Mediafile okay.. ", end = "")

    return True


def getimg(mediafile):

    """Acquires a numpy array from a video or image"""

    try:
        cap = cv2.VideoCapture(mediafile)
        _, img = cap.read()
    except:
        img = cv2.imread(mediafile)

    return img


def get_vid_params(mediafile):

    """Gets video parameters from file or video instance"""

    if type(mediafile) is str:
        if get_ext(mediafile) not in [".mov",".mp4",".avi"]:
            raise TypeError("File not a video..")
        mediafile = cv2.VideoCapture(mediafile)

    if not mediafile.read()[0]:
        raise RuntimeError("Video could not be read..")
    fps = int(mediafile.get(cv2.CAP_PROP_FPS))
    width = int(mediafile.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(mediafile.get(cv2.CAP_PROP_FRAME_HEIGHT))
    fcount = int(mediafile.get(cv2.CAP_PROP_FRAME_COUNT))

    return fps, width, height, fcount


def videowriter(filein, w, h, fps, resizeval = 1):

    """Creates a vidout instance using the opencv VideoWriter class"""

    ext = get_ext(filein)
    fileout = filein[:-len(ext)]+".mp4" if ext!="" else filein+".mp4"
    viddims = (w, h) if resizeval == 1 else (int(w*resizeval), int(h*resizeval))
    fourcc = cv2.VideoWriter_fourcc(*"mp4v")
    vidout = cv2.VideoWriter(fileout, fourcc, fps, viddims)

    return vidout


def safe_framecount(vidfile):

    """Saves video frame counter that counts frame-by-frame"""

    cap = cv2.VideoCapture(vidfile)
    vidlength = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

    count = 0
    while True:
        ret, frame = cap.read()
        if not ret:
            break
        count += 1

    print("video had", vidlength-count, "non-existing frames.. ", end = "")

    return count


def crop(image, pt1, pt2=None):

    """Crops image based on based on top left and bottom right corner"""

    if pt2 == None:
        pt2 = pt1[1]
        pt1 = pt1[0]
    cropped = image[pt1[1]:pt2[1], pt1[0]:pt2[0]]

    return cropped


def fourpt_transform(image, pts):

    """
    Perspective transform a section of an image based on four coordinates
    to obtain a top-down view
    """

    rect = sort_points(pts)
    (tl, tr, br, bl) = rect

    widthA = np.sqrt(((br[0] - bl[0]) ** 2) + ((br[1] - bl[1]) ** 2))
    widthB = np.sqrt(((tr[0] - tl[0]) ** 2) + ((tr[1] - tl[1]) ** 2))
    maxWidth = max(int(widthA), int(widthB))

    heightA = np.sqrt(((tr[0] - br[0]) ** 2) + ((tr[1] - br[1]) ** 2))
    heightB = np.sqrt(((tl[0] - bl[0]) ** 2) + ((tl[1] - bl[1]) ** 2))
    maxHeight = max(int(heightA), int(heightB))

    dst = np.array([[0, 0], [maxWidth - 1, 0],
                    [maxWidth - 1, maxHeight - 1],
                    [0, maxHeight - 1]], dtype = "float32")

    M = cv2.getPerspectiveTransform(rect, dst)
    warped = cv2.warpPerspective(image, M, (maxWidth, maxHeight))

    return warped


def checkroi(roi, resolution):

    """Make sure roi coordinates are within resolution"""

    x1 = max(roi[0][0],1)
    y1 = max(roi[0][1],1)
    x2 = min(roi[1][0],resolution[0])
    y2 = min(roi[1][1],resolution[1])

    return ((x1,y1),(x2,y2))


def zoom_to_roi(zoom, resolution):

    """Gets region of interest coordinates from x,y,w,h zoom parameters"""

    x1 = int(zoom[0] * resolution[0])
    x2 = int((zoom[0]+zoom[2]) * resolution[0])
    y1 = int(zoom[1] * resolution[1])
    y2 = int((zoom[1]+zoom[3]) * resolution[1])

    return ((x1,y1),(x2,y2))


def roi_to_zoom(roi, resolution):

    """Gets x,y,w,h zoom parameters from region of interest coordinates"""

    ((x1,y1),(x2,y2)) = roi
    z0 = round(x1 / resolution[0],2)
    z1 = round(y1 / resolution[1],2)
    z2 = round((x2-x1) / resolution[0],2)
    z3 = round((y2-y1) / resolution[1],2)

    return (z0, z1, z2, z3)


def picamconv(resolution, maxres = (1632, 1232)):

    """Adapts video resolution to work with raspberry pi camera"""

    width = min(closenr(resolution[0],32), maxres[0])
    height = min(closenr(resolution[1],16), maxres[1])

    return (width, height)


def fix_vidshape(res1,res2):

    """Compares two resolutions and get missing x and y coords"""

    xmin,ymin = 0,0
    xmult = (res2[0]/res1[0])
    ymult = (res2[1]/res1[1])
    if xmult > ymult:
        xmin = int((res2[0]-(res1[0]*ymult))/2)
    if ymult > xmult:
        ymin = int((res2[0]-(res1[0]*xmult))/2)

    return xmin, ymin


def newdims(img = None, resize = 1, dims = None):

    """Returns new dimensions of an image array based on resize value"""

    if dims is None:
        if img is None:
            print("No img or dims provided..")
            return
        else:
            dims = (img.shape[1],img.shape[0])

    width = int(dims[0] * resize)
    height = int(dims[1] * resize)

    return (width, height)


def imgresize(img, resize = 1, dims = None, back = False):

    """
    Returns resized image based on resizevalue or provided dimensions

    Parameters
    ----------
    img : numpy array
    resize : float, default = 1
        Multiplier for image size
    dims : tuple, default = None
        Dimensions of the to-be returned image
    back : bool, default = False
        If the inverse of the resize value should be used
    """

    if dims is None:
        resize = 1/resize if back else resize
        dims = newdims(img, resize)

    interpol = cv2.INTER_CUBIC if resize > 1 else cv2.INTER_AREA
    img = cv2.resize(img, dims, interpolation = interpol)

    return img


def add_transimg(bgimg, transimg, offsets):

    """
    Adds a semi-transparent (4-channel) image to a 3-channel background
    image. Images need to be arrays.
    """

    h, w, c = transimg.shape
    fix = np.zeros((h, w, 3), np.uint8)
    a = transimg[:, :, 3] / 255  #alpha
    o = offsets
    fix[:,:,0] = (1.-a)*bgimg[o[1]:o[1]+h, o[0]:o[0]+w, 0]+a*transimg[:,:,0]
    fix[:,:,1] = (1.-a)*bgimg[o[1]:o[1]+h, o[0]:o[0]+w, 1]+a*transimg[:,:,1]
    fix[:,:,2] = (1.-a)*bgimg[o[1]:o[1]+h, o[0]:o[0]+w, 2]+a*transimg[:,:,2]
    bgimg[o[1]:o[1]+h, o[0]:o[0]+w] = fix

    return bgimg



#! /usr/bin/env python
# Copyright (c) 2018 - 2019 Jolle Jolles <j.w.jolles@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import cv2
import numpy as np
import seaborn as sns

from pythutils.mathutils import sort_twoPoint


def textdims(text, size, thickness = 1):

    """Get uniform text dimensions for printing text with opencv"""

    tw, th = cv2.getTextSize(text, cv2.FONT_HERSHEY_SIMPLEX, size, thickness)[0]

    topy = size*1 if any(x in ["i","j"] for x in text) else 0
    boty = size*2 if any(x in "Q" for x in text) else 0
    boty = size*7 if any(x in ["g","j","y","p","q"] for x in text) else boty

    return (tw, th), topy, boty


class mouse_events:

    """Stores a series of coordinates related to mouse events"""

    def __init__(self):

        self.pos = (0,0)
        self.pts = []
        self.posDown = None
        self.posUp = None
        self.twoPoint = None
        self.drawing = False

    def draw(self, event, x, y, flags, param):

        if event == cv2.EVENT_MOUSEMOVE:
            self.pos = (x,y)

        if event == cv2.EVENT_LBUTTONDOWN:
            self.posDown = (x,y)
            self.pts.append((x,y))
            self.posUp = None

        if event == cv2.EVENT_LBUTTONUP:
            self.posUp = (x,y)
            self.twoPoint = sort_twoPoint((self.pts[-1], self.posUp))
            self.posDown = None

        self.drawing = True


def namedcols(colname = None, printlist = False, BRG = True):

    """Acquire RGB/BRG colors from a color name"""

    collist = {'black': (0, 0, 0),
             'navy': (0, 0, 128),
             'navyblue': (0, 0, 128),
             'darkblue': (0, 0, 139),
             'mediumblue': (0, 0, 205),
             'blue': (0, 0, 255),
             'darkgreen': (0, 100, 0),
             'green': (0, 128, 0),
             'darkcyan': (0, 139, 139),
             'deepskyblue': (0, 191, 255),
             'darkturquoise': (0, 206, 209),
             'mediumspringgreen': (0, 250, 154),
             'lime': (0, 255, 0),
             'springgreen': (0, 255, 127),
             'cyan': (0, 255, 255),
             'aqua': (0, 255, 255),
             'midnightblue': (25, 25, 112),
             'dodgerblue': (30, 144, 255),
             'lightseagreen': (32, 178, 170),
             'forestgreen': (34, 139, 34),
             'seagreen': (46, 139, 87),
             'darkslategray': (47, 79, 79),
             'darkslategrey': (47, 79, 79),
             'limegreen': (50, 205, 50),
             'mediumseagreen': (60, 179, 113),
             'turquoise': (64, 224, 208),
             'royalblue': (65, 105, 225),
             'steelblue': (70, 130, 180),
             'darkslateblue': (72, 61, 139),
             'mediumturquoise': (72, 209, 204),
             'indigo': (75, 0, 130),
             'darkolivegreen': (85, 107, 47),
             'cadetblue': (95, 158, 160),
             'cornflowerblue': (100, 149, 237),
             'mediumaquamarine': (102, 205, 170),
             'dimgray': (105, 105, 105),
             'dimgrey': (105, 105, 105),
             'slateblue': (106, 90, 205),
             'olivedrab': (107, 142, 35),
             'slategray': (112, 128, 144),
             'slategrey': (112, 128, 144),
             'lightslategray': (119, 136, 153),
             'lightslategrey': (119, 136, 153),
             'mediumslateblue': (123, 104, 238),
             'lawngreen': (124, 252, 0),
             'chartreuse': (127, 255, 0),
             'aquamarine': (127, 255, 212),
             'maroon': (128, 0, 0),
             'purple': (128, 0, 128),
             'olive': (128, 128, 0),
             'gray': (128, 128, 128),
             'grey': (128, 128, 128),
             'lightslateblue': (132, 112, 255),
             'skyblue': (135, 206, 235),
             'lightskyblue': (135, 206, 250),
             'blueviolet': (138, 43, 226),
             'darkred': (139, 0, 0),
             'darkmagenta': (139, 0, 139),
             'saddlebrown': (139, 69, 19),
             'darkseagreen': (143, 188, 143),
             'lightgreen': (144, 238, 144),
             'mediumpurple': (147, 112, 219),
             'darkviolet': (148, 0, 211),
             'palegreen': (152, 251, 152),
             'darkorchid': (153, 50, 204),
             'yellowgreen': (154, 205, 50),
             'sienna': (160, 82, 45),
             'brown': (165, 42, 42),
             'darkgray': (169, 169, 169),
             'darkgrey': (169, 169, 169),
             'lightblue': (173, 216, 230),
             'greenyellow': (173, 255, 47),
             'paleturquoise': (175, 238, 238),
             'lightsteelblue': (176, 196, 222),
             'powderblue': (176, 224, 230),
             'firebrick': (178, 34, 34),
             'darkgoldenrod': (184, 134, 11),
             'mediumorchid': (186, 85, 211),
             'rosybrown': (188, 143, 143),
             'darkkhaki': (189, 183, 107),
             'silver': (192, 192, 192),
             'mediumvioletred': (199, 21, 133),
             'indianred': (205, 92, 92),
             'peru': (205, 133, 63),
             'violetred': (208, 32, 144),
             'chocolate': (210, 105, 30),
             'tan': (210, 180, 140),
             'lightgray': (211, 211, 211),
             'lightgrey': (211, 211, 211),
             'thistle': (216, 191, 216),
             'orchid': (218, 112, 214),
             'goldenrod': (218, 165, 32),
             'palevioletred': (219, 112, 147),
             'crimson': (220, 20, 60),
             'gainsboro': (220, 220, 220),
             'plum': (221, 160, 221),
             'burlywood': (222, 184, 135),
             'lightcyan': (224, 255, 255),
             'lavender': (230, 230, 250),
             'darksalmon': (233, 150, 122),
             'violet': (238, 130, 238),
             'lightgoldenrod': (238, 221, 130),
             'palegoldenrod': (238, 232, 170),
             'lightcoral': (240, 128, 128),
             'khaki': (240, 230, 140),
             'aliceblue': (240, 248, 255),
             'honeydew': (240, 255, 240),
             'azure': (240, 255, 255),
             'sandybrown': (244, 164, 96),
             'wheat': (245, 222, 179),
             'beige': (245, 245, 220),
             'whitesmoke': (245, 245, 245),
             'mintcream': (245, 255, 250),
             'ghostwhite': (248, 248, 255),
             'salmon': (250, 128, 114),
             'antiquewhite': (250, 235, 215),
             'linen': (250, 240, 230),
             'lightgoldenrodyellow': (250, 250, 210),
             'oldlace': (253, 245, 230),
             'red': (255, 0, 0),
             'magenta': (255, 0, 255),
             'fuchsia': (255, 0, 255),
             'deeppink': (255, 20, 147),
             'orangered': (255, 69, 0),
             'tomato': (255, 99, 71),
             'hotpink': (255, 105, 180),
             'coral': (255, 127, 80),
             'darkorange': (255, 140, 0),
             'lightsalmon': (255, 160, 122),
             'orange': (255, 165, 0),
             'lightpink': (255, 182, 193),
             'pink': (255, 192, 203),
             'gold': (255, 215, 0),
             'peachpuff': (255, 218, 185),
             'navajowhite': (255, 222, 173),
             'moccasin': (255, 228, 181),
             'bisque': (255, 228, 196),
             'mistyrose': (255, 228, 225),
             'blanchedalmond': (255, 235, 205),
             'papayawhip': (255, 239, 213),
             'lavenderblush': (255, 240, 245),
             'seashell': (255, 245, 238),
             'cornsilk': (255, 248, 220),
             'lemonchiffon': (255, 250, 205),
             'floralwhite': (255, 250, 240),
             'snow': (255, 250, 250),
             'yellow': (255, 255, 0),
             'lightyellow': (255, 255, 224),
             'ivory': (255, 255, 240),
             'white': (255, 255, 255)}

    if printlist:
        print(collist)
    elif type(colname) == tuple and len(colname) == 3:
        return colname
    elif colname not in collist:
        print("colname does not exist..")
    else:
        col = collist[colname]
        col = (col[2],col[1],col[0]) if BRG else col
        return col


def uniqcols(colnr, colorspace="husl", rgb=True):

    """Gets a specified nr of colors, equally spaced on the color spectrum"""

    cols = sns.color_palette(colorspace, colnr)
    if rgb:
        cols = [tuple([(int(i*255)) for i in col])[::-1] for col in cols]
    return cols


def draw_text(img, text, loc = (0, 0), size = 1, col = "black", margin = 5,
              thickness = 1, bgcol = None, shadow = False):

    """
    Draw text on opencv image

    img : an image array
    text : the text to draw
    loc : the location of the text relative to topleft
    size : the size of the text
    col : the color of the text
    margin : the margin of the text
    thickness : the weight of the text
    bgcol : the potential background color of the text

    """

    col = namedcols(col)
    (tw, th), topy, boty = textdims(text, size)
    topy = topy + int(thickness/2)
    boty += int(thickness/2)

    if bgcol is not None:
        bgcol = namedcols(bgcol)
        topleftout = (loc[0], loc[1])
        botrightx = loc[0] + margin + tw + margin + 1
        botrighty = loc[1] + margin + th + topy + boty + margin + 1
        botrightout = (botrightx, botrighty)
        cv2.rectangle(img, topleftout, botrightout, bgcol, -1)

    botlefin = (int(loc[0]+margin), int(loc[1]+margin+th+topy))
    if shadow:
        cv2.putText(img, text, botlefin, cv2.FONT_HERSHEY_SIMPLEX, size,
                    (0,0,0), int(thickness*3), cv2.LINE_AA)
    cv2.putText(img, text, botlefin, cv2.FONT_HERSHEY_SIMPLEX, size,
                col, thickness, cv2.LINE_AA)


def draw_cross(img, pt1 = (1,1), pt2 = None, col = "white", thickness = 2):

    """Draws a cross"""

    pt2 = (img.shape[1],img.shape[0]) if pt2 == None else pt2
    cv2.line(img, pt1, (pt2[0], pt2[1]), namedcols(col), thickness)
    cv2.line(img, (pt2[0], pt1[0]), (pt1[1], pt2[1]), namedcols(col), thickness)


def draw_hcross(img, col="white", thickness=2, style="dashed"):

    """Draws a horizontal cross"""

    midTop = (int(img.shape[1]/2),1)
    midBot = (int(img.shape[1]/2),img.shape[0])
    midLeft = (1,int(img.shape[0]/2))
    midRight = (img.shape[1],int(img.shape[0]/2))
    draw_sliced_line(img, midTop, midBot, col, thickness, style)
    draw_sliced_line(img, midLeft, midRight, col, thickness, style)


def draw_crosshair(img, pt, radius = 5, col = "white"):

    """Draws a crosshair"""

    hline = (pt[0] - radius, pt[1]), (pt[0] + radius, pt[1])
    tline = (pt[0], pt[1] - radius), (pt[0], pt[1] + radius)
    cv2.line(img, hline[0], hline[1], namedcols(col), 1)
    cv2.line(img, tline[0], tline[1], namedcols(col), 1)


def draw_rectangle(img, pointer, rect, drawing = False, col = "red"):

    """Draws a rectangle with option to show it dynamically"""

    if drawing:
        cv2.rectangle(img, rect[0], pointer, namedcols(col), 2)

    else:
        cv2.rectangle(img, rect[0], rect[1], namedcols(col), 2)


def draw_bicircles(img, pt, resizeval = 1, col1 = "black", col2 = "white",
                   minsize = 3, maxsize = 15, stepsize = 3):

    """Draws a range of increasing bi-colored circles"""

    sizes = list(reversed(range(minsize, maxsize+stepsize, stepsize)))
    for i, size in enumerate(sizes):
        col = namedcols(col1) if i%2==0 else namedcols(col2)
        cv2.circle(img, pt, 0, col, int(size*resizeval))


def draw_sliced_line(img, pt1, pt2, color, thickness = 1, style = "dotted",
                     gap = 5):

    """Draw a dashed or dotted line on an image"""

    col = namedcols(color)
    dist =((pt1[0]-pt2[0])**2+(pt1[1]-pt2[1])**2)**.5
    pts= []

    for i in np.arange(0, dist, gap):
        r = i/dist
        x = int((pt1[0]*(1-r)+pt2[0]*r)+.5)
        y = int((pt1[1]*(1-r)+pt2[1]*r)+.5)
        pts.append((x,y))

    if style == "dotted":
        for pt in pts:
            cv2.circle(img, pt, thickness, col, -1)

    if style == "dashed":
        s = e = pts[0]
        for i,pt in enumerate(pts):
            s = e
            e = pt
            if i%2 == 1:
                cv2.line(img, s, e, col, thickness)


def draw_traj(img, coordlist = [], color = "green", thick_min = 8,
              thick_max = 13, opacity = 0.5):

    """Draws a semi-transparent polyline with decreasing width on an image"""

    col = namedcols(color)
    mask = img.copy()

    thicklist = np.linspace(thick_min, thick_max, len(coordlist))
    thicklist = (thicklist**4 / thick_max**4) * thick_max + 1
    thicklist = [max(i,1) for i in thicklist]

    for i in list(range(1, (len(coordlist) - 1))):
        thickness = int(thicklist[i])
        if coordlist[i] != coordlist[i] or coordlist[i-1] != coordlist[i-1]:
            continue
        elif None in sum((coordlist[i], coordlist[i-1]),()):
            continue

        cv2.line(mask, coordlist[i], coordlist[i-1], col, thickness)

    cv2.addWeighted(mask, opacity, img, 1-opacity, 0, img)



#! /usr/bin/env python
# Copyright (c) 2018 - 2019 Jolle Jolles <j.w.jolles@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import sys
import time
import datetime

def clock():

    """Simple running clock that prints on the same line"""

    while True:
        print(datetime.datetime.now().strftime("%H:%M:%S"), end="\r")
        sys.stdout.flush()
        time.sleep(1)


def now(timeformat = "date"):

    """Returns current date or time"""

    if timeformat == "date":
        return datetime.datetime.now().strftime("%y/%m/%d")

    elif timeformat == "time":
        return datetime.datetime.now().strftime("%H:%M:%S")

    else:
        print("No right time format provided..")



#! /usr/bin/env python
# Copyright (c) 2018 - 2019 Jolle Jolles <j.w.jolles@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages
import sys

exec(open("pythutils/__version__.py").read())

DESCRIPTION = """pythutils: a collection of utility functions for Python"""

DISTNAME = "pythutils"
MAINTAINER = "Jolle Jolles"
MAINTAINER_EMAIL = "j.w.jolles@gmail.com"
URL = "https://github.com/JolleJolles"
DOWNLOAD_URL = "https://github.com/JolleJolles/pythutils/archive/1.3.15.tar.gz"

with open("README.md") as f:
    readme = f.read()


if __name__ == "__main__":

    setup(name=DISTNAME,
          author=MAINTAINER,
          author_email=MAINTAINER_EMAIL,
          maintainer=MAINTAINER,
          maintainer_email=MAINTAINER_EMAIL,
          description=DESCRIPTION,
          long_description=readme,
          long_description_content_type="text/markdown",
          url=URL,
          download_url=DOWNLOAD_URL,
          version=__version__,
          license="License :: OSI Approved :: Apache Software License",
          platforms=["Windows", "Linux", "Mac OS-X"],
          packages=["pythutils"],
          install_requires=[
                     "numpy==1.16.5; python_version>='2' and python_version<'3'",
                     "numpy; python_version>='3'",
                     "pyyaml",
                     "objsize",
                     "seaborn"],
          classifiers=[
                     "Intended Audience :: Science/Research",
                     "Programming Language :: Python :: 2.7",
                     "Programming Language :: Python :: 3",
                     "License :: OSI Approved :: Apache Software License",
                     "Topic :: Scientific/Engineering :: Visualization",
                     "Topic :: Scientific/Engineering :: Image Recognition",
                     "Topic :: Scientific/Engineering :: Information Analysis",
                     "Topic :: Multimedia :: Video",
                     "Operating System :: POSIX",
                     "Operating System :: Unix",
                     "Operating System :: MacOS"],
          )



#! /usr/bin/env python
# Copyright (c) 2018 - 2019 Jolle Jolles <j.w.jolles@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import ast
import yaml

def move(file, newdir):

    """Moves a file to new location"""

    if not os.path.isfile(file):
        raise OSError("Provided file does not exist..")
    newdir = os.getcwd()+"/"+newdir if os.path.split(newdir)[0]=="" else newdir
    if not os.path.isdir(newdir):
        raise OSError("Provided newdir is no directory..")

    path, filename = os.path.split(file)
    print(path, filename)
    if path == "":
        path = os.getcwd()
    os.rename(path+"/"+filename, newdir+"/"+filename)


def listfiles(dir = ".", type = "", keepdir = False, keepext = True,
              nested = False):

    """
    Returns a list of (nested) files or directories

    dir: str; default="."
        Directory that should be checked
    type: str or tuple of strings; default=""
        Filetype for the files to be listed. If filetype is dir, it will return
    keepdir: bool; default=False
        If the original directory should be kept as part of the filename
    keepext: bool; defaul=True
        If the file extension should be kept in the filenames
    nested: bool; default=False
        If all nested files should be listed
    """

    if nested:
        outlist = []
        for root, dirs, files in os.walk(dir):
             for file in files:
                if file.endswith(type):
                    if keepdir:
                        outlist.append(os.path.join(root, file))
                    else:
                        outlist.append(file)

    else:
        if type == "dir":
            outlist = [i for i in os.listdir(dir) if os.path.isdir(os.path.join(dir, i))]

        else:
            outlist = [each for each in os.listdir(dir) if each.endswith(type)]
            outlist = [i for i in outlist if not i.startswith('.')]

        if keepdir:
            outlist = [dir + "/" + i  for i in outlist]

    outlist = sorted(outlist)

    if not keepext:
        outlist = [os.path.splitext(file)[0] for file in outlist]

    return outlist


def filechecker(indir = "", outdir = "", move = True, type=".h264",
                sleeptime = 2, function = None, functionparams = ("file")):

        """
        Repeatedly checks a directory for (new) files

        Parameters
        ===========
        indir : str, default = ""
            Directory with original files to monitor
        outdir : str, default = ""
            Directory with new files to monitor
        type : str, default = ".h264"
            The type of file to monitor
        sleeptime : int, default = 2
            Time in seconds between subsequent checks of file folder
        function :
            function to use with the files to check
        functionparams :
            parameters for the function to use
        """

        global interrupted
        interrupted = False
        def keythread():
            global interrupted
            input()
            interrupted = True
        Thread(target=keythread).start()

        if function is None:
            raise Inputerror("No function is provided")
        indir = os.getcwd() if indir == "" else indir
        if not os.path.exists(indir):
            raise OSError("in-directory does not exist..")
        outdir = indir if outdir == "" else outdir
        if not os.path.exists(outdir):
            raise OSError("out-directory does not exist..")

        rootdir = os.getcwd()
        os.chdir(indir)

        while not interrupted:
            files = listfiles(indir, type, keepdir = False)
            if not move:
                old = listfiles(indir, type, keepext = False)
                new = listfiles(outdir, type, keepext = False)
                files = [files[i] for i,file in enumerate(old) if file not in new]
            for file in files:
                if "file" == functionparams:
                    function(file)
                elif "file" in functionparams:
                    function(file, *functionparams[1:])
                else:
                    function(*functionparams)

            lineprint("No files found, rechecking in "+str(sleeptime)+"s..")
            time.sleep(sleeptime)

        os.chdir(rootdir)
        lineprint("Filechecking stopped..")


def commonpref(pathlist = None, remove = False):

    """Given a list of paths, returns the longest common leading component"""

    if pathlist is None: return ""
    s1 = min(pathlist)
    s2 = max(pathlist)
    for i, c in enumerate(s1):
        if c != s2[i]:
            commcomp = s1[:i]
            break

    if remove:
        for each in range(len(pathlist)):
            pathlist[each] = pathlist[each].split(commcomp, 1)[-1]
        return pathlist

    else:
        return commcomp


def get_ext(filename):

    """Returns file extension in lower case"""

    return os.path.splitext(str(filename))[-1].lower()


def loadyml(filename, value = None, add = True):

    """Loads value from .yml file and returns literal"""

    if os.path.exists(filename):
        with open(filename) as f:
            newvalue = yaml.load(f)
        if value is not None:
            newvalue = newvalue + value if add else value
    else:
        newvalue = value
    newvalue = ast.literal_eval(str(newvalue))

    return newvalue


def loadh5data(filename, dataset = "data"):

    import h5py
    import pandas
    h5file = h5py.File(filename, 'r')
    dataset = pandas.DataFrame(h5file[dataset][:])
    h5file.close()

    return dataset


def name(filename, ext = "", action = "newfile"):

    """
    Gets the name for a file with required extension, and will either overwrite
    the existing file or generate a new file with a sequence.
    """

    dirname, filename = os.path.split(filename)
    dirname = '.' if dirname == '' else dirname
    filename, ext = os.path.splitext(filename)
    names = [x for x in os.listdir(dirname) if x.startswith(filename)]
    names = [x for x in names if os.path.splitext(x)[1]==ext]

    if len(names) == 0 or action == "append":
        return filename+ext

    elif action == "overwrite":
        if os.path.exists(filename+ext):
            os.remove(filename+ext)
        return filename+ext

    elif action == "newfile":
        names = [os.path.splitext(x)[0] for x in names]
        suffixes = [x.replace(filename, '') for x in names]
        suffixes = [int(x[1]) for x in suffixes if x.startswith('_')]
        suffix = 2 if len(suffixes)==0 else max(suffixes)+1
        return '%s_%d%s' % (filename, suffix, ext)



#! /usr/bin/env python
# Copyright (c) 2018 - 2019 Jolle Jolles <j.w.jolles@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import numpy as np

def _list_to_coords(list):

    coords = [(int(i),int(j)) for i,j in list if i==i]
    loclist = [c for c,i in enumerate(list) if i[0] == i[0]]

    return coords, loclist


def create_emptydf(cols = ["x","y","fx","fy"], ids = [1], first = 1, last = None):

    """
    Creates an emtpy pandas dataframe, aimed to hold coordinate data

    cols : a list of columns
    ids : a list of animal ids
    first : start frame
    last : last frame
    """

    try:
        framerange = range(first, last + 1)
    except TypeError:
        raise TypeError("No last value provided..")

    colnames = ["frame","id"] + cols
    emptycols = list(np.repeat(np.nan, len(cols)))

    for i, id in enumerate(ids):
        sub = pd.DataFrame([[frame, id] + emptycols for frame in framerange],
              columns = colnames)
        data = sub if i == 0 else pd.concat([data, sub])

    data = data.sort_values(["frame"])
    data.index = range(0, data.index.size, 1)

    return data


def pd_to_coords(pdat, loc = None, array = False, columns = ["x","y"],
                 multiplier = 1):

    """
    Returns either a single coordinate of integers or a list or an array of
    arrays with coordinates with a list of frames
    """

    if loc != None:
        coords = _list_to_coords([list(pdat.loc[loc, columns])])[0]
        if len(coords) == 0:
            return None
        else:
            c = coords[0]
            return (int(c[0]*multiplier),int(c[1]*multiplier))

    else:
        coords, loclist = _list_to_coords(list(zip(pdat[columns[0]],
                                                   pdat[columns[1]])))
        framelist = [pdat.loc[i,"frame"] for i in loclist]
        coords = [(int(c[0]*multiplier),int(c[1]*multiplier)) for c in coords]
        if array:
            coords = [[[i] for i in coords]]
            coords = np.array(coords, np.int32).reshape((-1,1,2))

        return coords, framelist


def dfchange(df1, df2):

    """Determines the differences between two pandas dataframes"""

    dfchanges = df2.loc[df2[~df2.isin(df1)].dropna(how="all").index,]
    nchanges = dfchanges.shape[0]

    return dfchanges, nchanges


def to_query(cols, vals):

    """Returns a query from a list of columns and values"""

    vals = ["'"+val+"'" if isinstance(val,str) else val for val in vals]
    query = ['{} == {}'.format(cols[i],vals[i]) for i,_ in enumerate(cols)]
    query = ' and '.join(query)

    return query


def contour_to_tuple(contour):

    """Convert contour to tuple"""

    conttuple = [(row[0][0], row[0][1]) for row_idx, row in enumerate(contour)]

    return conttuple
